/*
 * Copyright © 2012-2013 Intel Corporation
 *
 * Authors: Rodrigo Moya <rodrigo.moya@collabora.co.uk>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib/gi18n.h>
#include "config.h"
#include "diagnostics-view.h"
#include "main-window.h"
#include "package-manager.h"
#include "transaction-view.h"

#define WAITING " ⏳ "

G_DEFINE_TYPE (MainWindow, main_window, GTK_TYPE_DIALOG)

struct _MainWindowPrivate {
  /* Widgets */
  GtkWidget *notebook;
  GtkWidget *intro_view;
  GtkWidget *diagnostics_page;
  GtkWidget *transaction_page;
  GtkWidget *back_button;
  GtkWidget *close_button;
  GtkWidget *uninstall_button;
  GtkWidget *next_button;
  struct {
    GtkWidget *status;
    GtkWidget *message;
    GtkWidget *percent;
    GtkWidget *spinner;
  } progress;
};

static void
main_window_finalize (GObject *object)
{
  MainWindow *win = MAIN_WINDOW (object);

  if (win->priv != NULL)
    {
      g_free (win->priv);
    }

  G_OBJECT_CLASS (main_window_parent_class)->finalize (object);
}

static void
main_window_class_init (MainWindowClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->finalize = main_window_finalize;
}

static void
on_diagnostics_finished (DiagnosticsView *view, gboolean success, gpointer user_data)
{
  MainWindow *win = MAIN_WINDOW (user_data);

  g_debug ("%s/%s: Diagnostics finished %s", __FILE__, __FUNCTION__, success ? "successfully" : "with an error");

  if (success)
    {
      PackageManager *pm = package_manager_get ();

      /* If the package manager is configured, or we have some non-vanilla *
       * packages, we should offer the option to downgrade:                */
      if (package_manager_is_setup (pm) ||
          package_manager_is_system_upgraded (pm, view))
        {
#if ENABLE_UNINSTALL
          gtk_widget_show (win->priv->uninstall_button);
#endif
          gtk_button_set_label (GTK_BUTTON (win->priv->next_button), _("Install"));
        }

      gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_APPLY, TRUE);
    }

  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_REJECT, TRUE);
  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_CLOSE, TRUE);
}

static void
on_transaction_finished (TransactionView *view, gboolean success, gpointer user_data)
{
  MainWindow *win = MAIN_WINDOW (user_data);

  g_debug ("%s/%s: Package transaction finished %s", __FILE__, __FUNCTION__, success ? "successfully" : "with an error");

  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_REJECT, TRUE);
  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_CLOSE, TRUE);
}

static void
on_dialog_response (GtkDialog *dialog, gint response_id, gpointer user_data)
{
  MainWindow *win = MAIN_WINDOW (dialog);

  switch (response_id)
    {
    case GTK_RESPONSE_REJECT:
      switch (gtk_notebook_get_current_page (GTK_NOTEBOOK (win->priv->notebook)))
	{
	case 1: /* Diagnostics page */
	  gtk_notebook_set_current_page (GTK_NOTEBOOK (win->priv->notebook), 0);
	  gtk_widget_hide (win->priv->back_button);
	  gtk_widget_hide (win->priv->uninstall_button);
	  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_APPLY, TRUE);
	  gtk_button_set_label (GTK_BUTTON (win->priv->next_button), _("Next"));
	  break;
	case 2: /* Transaction page */
	  gtk_notebook_set_current_page (GTK_NOTEBOOK (win->priv->notebook), 1);
	  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_APPLY, TRUE);
	  if (package_manager_is_setup (package_manager_get ()))
	    {
#if ENABLE_UNINSTALL
          gtk_widget_show (win->priv->uninstall_button);
#endif
          gtk_button_set_label (GTK_BUTTON (win->priv->next_button), _("Install"));
	    }
	  else
	    {
	      gtk_widget_hide (win->priv->uninstall_button);
	      gtk_button_set_label (GTK_BUTTON (win->priv->next_button), _("Next"));
	    }
	  break;
	}
      break;
    case GTK_RESPONSE_CLOSE:
      gtk_widget_destroy (GTK_WIDGET (win));
      break;
    case GTK_RESPONSE_CANCEL:
      gtk_notebook_set_current_page (GTK_NOTEBOOK (win->priv->notebook), 2);
      gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_REJECT, FALSE);
      gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_CLOSE, FALSE);
      gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_APPLY, FALSE);
      gtk_widget_hide (win->priv->uninstall_button);
      gtk_button_set_label (GTK_BUTTON (win->priv->next_button), _("Next"));

      g_signal_connect (win->priv->transaction_page, "finished",
			G_CALLBACK (on_transaction_finished), win);
      transaction_view_start (TRANSACTION_VIEW (win->priv->transaction_page), TRUE);
      break;
    case GTK_RESPONSE_APPLY:
      switch (gtk_notebook_get_current_page (GTK_NOTEBOOK (win->priv->notebook)))
	{
	case 0: /* Introduction page */
	  gtk_widget_show (win->priv->back_button);
	  gtk_notebook_set_current_page (GTK_NOTEBOOK (win->priv->notebook), 1);
	  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_REJECT, FALSE);
	  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_CLOSE, FALSE);
	  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_APPLY, FALSE);

	  g_signal_connect (win->priv->diagnostics_page, "finished",
			    G_CALLBACK (on_diagnostics_finished), win);
	  diagnostics_view_start (DIAGNOSTICS_VIEW (win->priv->diagnostics_page));
	  break;
	case 1: /* Diagnostics page */
	  gtk_notebook_set_current_page (GTK_NOTEBOOK (win->priv->notebook), 2);
	  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_REJECT, FALSE);
	  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_CLOSE, FALSE);
	  gtk_dialog_set_response_sensitive (GTK_DIALOG (win), GTK_RESPONSE_APPLY, FALSE);
	  gtk_widget_hide (win->priv->uninstall_button);
	  gtk_button_set_label (GTK_BUTTON (win->priv->next_button), _("Next"));

	  g_signal_connect (win->priv->transaction_page, "finished",
			    G_CALLBACK (on_transaction_finished), win);
	  transaction_view_start (TRANSACTION_VIEW (win->priv->transaction_page), FALSE);
	  break;
	}
      break;
    }
}

static void
progress_widgets_init (MainWindow *win, GtkBox *box)
{
  GtkWidget *w;
  GtkLabel *l;

  /* create the progress widget(s) */
  w = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 1);
  gtk_widget_set_halign (w, GTK_ALIGN_START);
  gtk_widget_set_valign (w, GTK_ALIGN_CENTER);

  /* the short "current action" widget */
  win->priv->progress.status = gtk_label_new (" ");
  l = GTK_LABEL (win->priv->progress.status);
  gtk_label_set_ellipsize(l, PANGO_ELLIPSIZE_MIDDLE);
  gtk_label_set_width_chars (l, 20);
  gtk_label_set_max_width_chars(l, 20);
  gtk_misc_set_alignment (GTK_MISC (l), 0.0, 0.0);
  gtk_box_pack_start(GTK_BOX (w), win->priv->progress.status, FALSE, FALSE, 1);
  gtk_widget_show (win->priv->progress.status);

  /* The longer detailed-message widget */
  win->priv->progress.message = gtk_label_new (" ");
  l = GTK_LABEL (win->priv->progress.message);
  gtk_box_pack_start(GTK_BOX (w), win->priv->progress.message, FALSE, FALSE, 1);
  gtk_label_set_ellipsize (l, PANGO_ELLIPSIZE_MIDDLE);
  gtk_label_set_width_chars (l, 40);
  gtk_label_set_max_width_chars (l, 40);
  gtk_misc_set_alignment (GTK_MISC (l), 0.0, 0.0);
  gtk_widget_show (win->priv->progress.message);

  /* The overall %age widget */
  win->priv->progress.percent = gtk_label_new (" ");
  l = GTK_LABEL (win->priv->progress.percent);
  gtk_label_set_width_chars(l, 5);
  gtk_label_set_max_width_chars(l, 5);
  gtk_misc_set_alignment (GTK_MISC (l), 0.0, 0.0);
  gtk_box_pack_start (GTK_BOX (w), win->priv->progress.percent, FALSE, FALSE, 1);
  gtk_widget_show (win->priv->progress.percent);

  /* the progress spinner */
  win->priv->progress.spinner = gtk_spinner_new ();
  gtk_box_pack_end (GTK_BOX (w), win->priv->progress.spinner, TRUE, FALSE, 1);

  gtk_widget_show (w);
  gtk_box_pack_start (box, w, TRUE, TRUE, 3);
}

static void
main_window_init (MainWindow *win)
{
  GtkWidget *vbox;
  GtkWidget *box, *w;

  win->priv = g_new0 (MainWindowPrivate, 1);

  vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 3);
  box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 3);

  gtk_widget_show (box);
  gtk_widget_show (vbox);

  gtk_box_pack_start (GTK_BOX (gtk_dialog_get_content_area (GTK_DIALOG (win))), vbox, TRUE, TRUE, 3);

  progress_widgets_init (win, GTK_BOX (vbox));
  gtk_box_pack_start (GTK_BOX (vbox), box, TRUE, TRUE, 1);

  w = gtk_image_new_from_file (IMAGESDIR "/logo.png");
  gtk_widget_show (w);
  gtk_box_pack_start (GTK_BOX (box), w, FALSE, TRUE, 3);

  /* Create the notebook that will provide the different pages */
  win->priv->notebook = gtk_notebook_new ();
  gtk_container_set_border_width (GTK_CONTAINER (win->priv->notebook), 12);
  gtk_notebook_set_show_tabs (GTK_NOTEBOOK (win->priv->notebook), FALSE);
  gtk_notebook_set_show_border (GTK_NOTEBOOK (win->priv->notebook), FALSE);
  gtk_widget_show (win->priv->notebook);
  gtk_box_pack_start (GTK_BOX (box), win->priv->notebook, TRUE, TRUE, 3);

  /* Add Introduction view */
  win->priv->intro_view = gtk_label_new (_("<big><b>Intel Driver Update Utility for Linux Systems</b></big>\n\n"
					   "<b>Automatically identify and find software updates.</b>\n\n"
					   "This program allows you to install the latest graphics stack packages "
					   "from Intel. It will find and setup the repositories needed for getting "
					   "your system up-to-date with the latest graphics drivers from Intel."));
  gtk_label_set_use_markup (GTK_LABEL (win->priv->intro_view), TRUE);
  gtk_label_set_line_wrap (GTK_LABEL (win->priv->intro_view), TRUE);
  gtk_misc_set_alignment (GTK_MISC (win->priv->intro_view), 0.0, 0.0);
  gtk_notebook_append_page (GTK_NOTEBOOK (win->priv->notebook), win->priv->intro_view, NULL);
  gtk_widget_show (win->priv->intro_view);

  /* Add Diagnostics view */
  win->priv->diagnostics_page = diagnostics_view_new ();
  gtk_notebook_append_page (GTK_NOTEBOOK (win->priv->notebook), win->priv->diagnostics_page, NULL);
  gtk_widget_show (win->priv->diagnostics_page);

  /* Add Transaction view */
  win->priv->transaction_page = transaction_view_new ();
  gtk_notebook_append_page (GTK_NOTEBOOK (win->priv->notebook), win->priv->transaction_page, NULL);
  gtk_widget_show (win->priv->transaction_page);

  /* Add buttons */
  win->priv->back_button = gtk_dialog_add_button (GTK_DIALOG (win), GTK_STOCK_GO_BACK, GTK_RESPONSE_REJECT);
  gtk_widget_hide (win->priv->back_button);
  win->priv->close_button = gtk_dialog_add_button (GTK_DIALOG (win), GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE);
  win->priv->uninstall_button = gtk_dialog_add_button (GTK_DIALOG (win), _("Uninstall"), GTK_RESPONSE_CANCEL);
  gtk_widget_hide (win->priv->uninstall_button);
  win->priv->next_button = gtk_dialog_add_button (GTK_DIALOG (win), _("Next"), GTK_RESPONSE_APPLY);

  g_signal_connect (win, "response", G_CALLBACK (on_dialog_response), NULL);
}

GtkWindow *
main_window_new (void)
{
  return g_object_new (TYPE_MAIN_WINDOW, NULL);
}

static void copy_str (gchar *dst, const gchar *src, gsize len)
{
  strncpy (dst, src, len);
  dst[len - 1] = '\0';
}

void
main_window_progress (MainWindow *win,
                      const gchar *action,
                      const gchar *msg,
                      gint percent,
                      gboolean spin)
{
  static gchar last_action[255] = "";
  static gint  last_percent = -1;
  static gint  stickiness = 0;
  gboolean repeat = g_strcmp0 (action, last_action) == 0;
  gboolean stuck = FALSE;

  if (action)
    copy_str (last_action, action, 255);

  // still going, same action as before, percent didn't change:
  if (spin && repeat && last_percent == percent)
    stuck = (stickiness++ >= 2);
  else
    stickiness = 0;

  last_percent = percent;

  if (win == NULL || win->priv == NULL)
    return;

  if (action != NULL && win->priv->progress.status != NULL)
    {
      GtkLabel *a = GTK_LABEL (win->priv->progress.status);
      gtk_label_set_text(a, action);
    }

  if (msg != NULL && win->priv->progress.message != NULL)
    {
      GtkLabel *m = GTK_LABEL (win->priv->progress.message);
      gtk_label_set_text(m, msg);
    }

  if (win->priv->progress.spinner != NULL)
    {
      GtkSpinner *s = GTK_SPINNER(win->priv->progress.spinner);

      if (spin)
        {
          gtk_spinner_start (s);
          gtk_widget_show (GTK_WIDGET (s));
        }
      else
        {
          gtk_spinner_stop (s);
          gtk_widget_hide (GTK_WIDGET (s));
        }
    }

  if (percent != -1 &&
      win->priv->progress.percent != NULL &&
      win->priv->progress.spinner != NULL)
    {
      GtkLabel *p = GTK_LABEL (win->priv->progress.percent);

      if (percent > 100)
        percent = 100;

      if (stuck)
        {
          gtk_label_set_text (p, WAITING);
        }
      else if (percent >= 0)
        {
          gchar *text = g_strdup_printf ("%3d %%", percent);

          gtk_label_set_text (p, text);
          g_free (text);
        }
      else
        {
          gtk_label_set_text (p, WAITING);
        }
    }

    gboolean spinning = FALSE;
    g_object_get (win->priv->progress.spinner, "active", &spinning, NULL);
    g_debug ("🕛 %s : %s  [ %s ] %s",
             gtk_label_get_text (GTK_LABEL (win->priv->progress.status )),
             gtk_label_get_text (GTK_LABEL (win->priv->progress.message)),
             gtk_label_get_text (GTK_LABEL (win->priv->progress.percent)),
             spinning ? "⏲" : "◦");
}
