/*
 * Copyright © 2012-2013 Intel Corporation
 *
 * Authors: Rodrigo Moya <rodrigo.moya@collabora.co.uk>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib/gi18n.h>
#include <gio/gio.h>
#include <polkit/polkit.h>
#include "package-manager.h"
#include "package-manager-fedora.h"
#include "package-manager-ubuntu.h"
#include "utils.h"
#include "marshal.h"

G_DEFINE_TYPE (PackageManager, package_manager, G_TYPE_OBJECT)

enum {
  ERROR_SIGNAL,
  PROGRESS_SIGNAL,
  SETUP_FINISHED_SIGNAL,
  TRANSACTION_FINISHED_SIGNAL,
  CLEANUP_FINISHED_SIGNAL,
  LAST_SIGNAL
};
static guint pm_signals[LAST_SIGNAL];

static GHashTable *pm_list = NULL;

static void
package_manager_finalize (GObject *object)
{
  G_OBJECT_CLASS (package_manager_parent_class)->finalize (object);
}

static void
package_manager_class_init (PackageManagerClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->finalize = package_manager_finalize;
  klass->is_setup = NULL;
  klass->is_distro_supported = NULL;
  klass->setup = NULL;
  klass->run_transaction = NULL;
  klass->clean = NULL;
  klass->is_system_upgraded = NULL;

  /* Install signals */
  pm_signals[ERROR_SIGNAL] = g_signal_new ("error",
					   G_OBJECT_CLASS_TYPE (object_class),
					   G_SIGNAL_RUN_LAST,
					   G_STRUCT_OFFSET (PackageManagerClass, error),
					   NULL, NULL,
					   g_cclosure_marshal_VOID__STRING,
					   G_TYPE_NONE, 1, G_TYPE_STRING);
  pm_signals[PROGRESS_SIGNAL] = g_signal_new ("progress",
					      G_OBJECT_CLASS_TYPE (object_class),
					      G_SIGNAL_RUN_LAST,
					      G_STRUCT_OFFSET (PackageManagerClass, progress),
					      NULL, NULL,
					      g_cclosure_marshal_VOID__STRING,
					      G_TYPE_NONE, 1, G_TYPE_STRING);
  pm_signals[SETUP_FINISHED_SIGNAL] = g_signal_new ("setup-finished",
						    G_OBJECT_CLASS_TYPE (object_class),
						    G_SIGNAL_RUN_LAST,
						    G_STRUCT_OFFSET (PackageManagerClass, setup_finished),
						    NULL, NULL,
						    g_cclosure_marshal_VOID__VOID,
						    G_TYPE_NONE, 0);
  pm_signals[TRANSACTION_FINISHED_SIGNAL] = g_signal_new ("transaction-finished",
							  G_OBJECT_CLASS_TYPE (object_class),
							  G_SIGNAL_RUN_LAST,
							  G_STRUCT_OFFSET (PackageManagerClass, transaction_finished),
							  NULL, NULL,
							  g_cclosure_marshal_VOID__VOID,
							  G_TYPE_NONE, 0);
  pm_signals[CLEANUP_FINISHED_SIGNAL] = g_signal_new ("cleanup-finished",
						      G_OBJECT_CLASS_TYPE (object_class),
						      G_SIGNAL_RUN_LAST,
						      G_STRUCT_OFFSET (PackageManagerClass, cleanup_finished),
						      NULL, NULL,
						      g_cclosure_marshal_VOID__VOID,
						      G_TYPE_NONE, 0);
}

static void
package_manager_init (PackageManager *manager)
{
}

PackageManager *
package_manager_get (void)
{
  PackageManager *manager = NULL;
  Distribution *distro;

  /* Determine on which distro we are running */
  distro = get_distro_information ();
  if (!distro)
    return NULL;

  /* See if we already have a distro package manager */
  if (pm_list != NULL)
    {
      manager = g_hash_table_lookup (pm_list, distro->distributor);
    }

  if (!manager)
    {
      if (g_strcmp0 (distro->distributor, "Ubuntu") == 0)
	manager = package_manager_ubuntu_new ();
      else if (g_strcmp0 (distro->distributor, "Fedora") == 0)
	manager = package_manager_fedora_new ();

      if (manager != NULL)
	{
	  if (!pm_list)
	    pm_list = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
	  g_hash_table_insert (pm_list, g_strdup (distro->distributor), manager);
	}
    }

  return manager;
}

/* This is intended to be called _only_ when the 3rd party repo(s) are NOT   *
 * configured and we want to check to see if we have upgraded packages, ie   *
 * we are partially upgraded and we want to let the user head either way:    *
 * It will not return sensible results if the 3rd party repos are configured */
gboolean
package_manager_is_system_upgraded (PackageManager *manager, DiagnosticsView *v)
{
  gboolean upgraded = FALSE;

  g_return_val_if_fail (IS_PACKAGE_MANAGER (manager), FALSE);

  PackageManagerClass *klass = PACKAGE_MANAGER_GET_CLASS (manager);

  update_progress_details ("Status-check", "Checking upgrade status", 0, 1);

  if (klass->is_system_upgraded != NULL)
      upgraded = klass->is_system_upgraded (manager, v);

  update_progress_details (NULL, "Complete", 101, 0);

  return upgraded;
}

gboolean
package_manager_is_setup (PackageManager *manager)
{
  gboolean pmsetup = FALSE;

  g_return_val_if_fail (IS_PACKAGE_MANAGER (manager), FALSE);

  update_progress_details ("Status-check",
                           "Checking Repository Configuration", 0, 1);

  if (PACKAGE_MANAGER_GET_CLASS (manager)->is_setup != NULL)
    pmsetup = PACKAGE_MANAGER_GET_CLASS (manager)->is_setup (manager);

  update_progress_details (NULL, "Complete", 101, 0);

  return pmsetup;
}

gboolean
package_manager_is_distro_supported (PackageManager *manager)
{
  gboolean supported = FALSE;
  Configuration *cfg = NULL;
  PackageManagerClass *klass = PACKAGE_MANAGER_GET_CLASS (manager);

  g_return_val_if_fail (IS_PACKAGE_MANAGER (manager), FALSE);

  update_progress_details ("Configuration", "Loading …", 0, 1);

  cfg = get_configuration ();

  update_progress_details (NULL, "Loaded", 80, 1);

  if (!cfg)
    {
      g_signal_emit_by_name (manager, "error", "Unsupported distribution");
      return FALSE;
    }

  if (cfg->errors && cfg->errors->len > 0)
    {
      guint i;

      for (i = 0; i < cfg->errors->len; i++)
        {
          gchar *msg = g_ptr_array_index (cfg->errors, i);

          g_signal_emit_by_name (manager, "error", msg);
        }

      return FALSE;
    }

  update_progress_details (NULL, "Testing …", 80, 1);

  if (klass->is_distro_supported != NULL)
    supported = klass->is_distro_supported (manager, cfg);

  update_progress_details (NULL, "Complete", 101, 0);

  return supported;
}

gboolean
package_manager_setup (PackageManager *manager)
{
  gboolean is_setup = FALSE;

  g_return_val_if_fail (IS_PACKAGE_MANAGER (manager), FALSE);

  update_progress_details ("Setup", "Configuring Repositories …", 0, 1);

  if (PACKAGE_MANAGER_GET_CLASS (manager)->setup != NULL)
    is_setup = PACKAGE_MANAGER_GET_CLASS (manager)->setup (manager);

  update_progress_details (NULL, "Complete", 101, 0);

  return is_setup;
}

gboolean
package_manager_run_transaction (PackageManager *manager)
{
  g_return_val_if_fail (IS_PACKAGE_MANAGER (manager), FALSE);

  if (PACKAGE_MANAGER_GET_CLASS (manager)->run_transaction != NULL)
    return PACKAGE_MANAGER_GET_CLASS (manager)->run_transaction (manager);

  return FALSE;
}

gboolean
package_manager_clean (PackageManager *manager)
{
  g_return_val_if_fail (IS_PACKAGE_MANAGER (manager), FALSE);

  if (PACKAGE_MANAGER_GET_CLASS (manager)->clean != NULL)
    return PACKAGE_MANAGER_GET_CLASS (manager)->clean (manager);

  return FALSE;
}

gboolean
package_manager_check_pk_permission (PackageManager *manager, const gchar *action_id)
{
  GPermission *permission;
  gboolean result = TRUE;

  g_debug ("polkit: Fetching %s permissions...", action_id);
  permission = polkit_permission_new_sync (action_id, NULL, NULL, NULL);
  if (!permission)
    {
      gchar *msg = g_strdup_printf (_("Your system does not have the '%s' PolicyKit files installed. Please check your installation"), action_id);
      g_signal_emit_by_name (manager, "error", msg);
      g_debug ("polkit: %s; %s", action_id, msg);
      g_free (msg);

      return FALSE;
    }

  if (!g_permission_get_allowed (permission))
    {
      GError *error = NULL;

      if (!g_permission_get_can_acquire (permission))
	{
	  g_signal_emit_by_name (manager, "error", _("You don't have permissions to run this operation"));
      g_debug ("polkit: %s; %s", action_id, "permission cannot be acquired");
	  result = FALSE;
	}
      else if (!g_permission_acquire (permission, NULL, &error))
	{
	  gchar *msg = g_strdup_printf (_("Could not acquire permissions: %s"), error->message);

	  g_signal_emit_by_name (manager, "error", msg);
      g_debug ("polkit: %s; %s", action_id, msg);
	  g_free (msg);
	  g_error_free (error);

	  result = FALSE;
	}
    }

  g_object_unref (permission);

  return result;
}
